/*
 * Server binding/registration.
 *
 * Copyright (C) 2018, 2019, 2020, 2021, 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/env.h>
#include <l4/sys/factory.h>
#include <l4/sys/kernel_object.h>
#include <l4/sys/rcv_endpoint.h>
#include <l4/sys/thread.h>
#include <l4/util/util.h>

#include <stdlib.h>

#include "cap_alloc.h"
#include "irq.h"
#include "server.h"



/* Associate an IRQ with the IPC gate in the main thread to handle gate deletion
   notifications. */

long ipc_server_apply_irq(l4_cap_idx_t cap, l4_cap_idx_t *irq)
{
  return ipc_server_apply_irq_for_thread(cap, irq, l4re_env()->main_thread);
}

/* Associate an IRQ with the IPC gate in the given thread to handle gate deletion
   notifications. */

long ipc_server_apply_irq_for_thread(l4_cap_idx_t cap, l4_cap_idx_t *irq, l4_cap_idx_t thread)
{
  /* Create an IRQ for the gate. */

  long err = ipc_create_irq(irq);

  if (err)
    return err;

  /* Bind the IRQ to the thread, presenting the label if it provides the
     incoming message. */

  err = l4_error(l4_rcv_ep_bind_thread(*irq, thread, (l4_umword_t) *irq));

  if (err)
  {
    ipc_cap_free_um(*irq);
    return err;
  }

  /* Register for notifications. The thread associated with the gate seems to be
     needed. */

  err = l4_error(l4_thread_register_del_irq(thread, *irq));

  if (err)
  {
    ipc_cap_free_um(*irq);
    return err;
  }

  /* Decrease the reference count on the server capability (IPC gate). */

  return l4_error(l4_kobject_dec_refcnt(cap, 1));
}



/* Bind the main thread to the named capability, employing an identifier for
   the IPC gate and returning the capability index. */

long ipc_server_bind(const char *name, l4_umword_t id, l4_cap_idx_t *server)
{
  /* Obtain a reference to the device. */

  *server = l4re_env_get_cap(name);

  if (l4_is_invalid_cap(*server))
    return -L4_ENOENT;

  /* Bind to an IPC gate with the label identifying the server as message
     destination. */

  return l4_error(l4_rcv_ep_bind_thread(*server, l4re_env()->main_thread, id));
}

/* Initialise a new server capability, binding it to the main thread. */

long ipc_server_new(l4_cap_idx_t *cap, void *obj)
{
  return ipc_server_new_for_thread(cap, obj, l4re_env()->main_thread);
}

/* Initialise a new server capability, binding it to the given thread. */

long ipc_server_new_for_thread(l4_cap_idx_t *cap, void *obj, l4_cap_idx_t thread)
{
  l4_umword_t label = (l4_umword_t) obj;

  /* Create an IPC gate, presenting the resource's label if this gate provides
     the message. */

  return ipc_server_new_gate_for_thread(cap, thread, label);
}

/* Create an IPC gate, exporting via a new capability an object with the given
   identifier. */

long ipc_server_new_gate(l4_cap_idx_t *ref, l4_umword_t id)
{
  return ipc_server_new_gate_for_thread(ref, l4re_env()->main_thread, id);
}

/* Create an IPC gate associated with a specific thread. */

long ipc_server_new_gate_for_thread(l4_cap_idx_t *ref, l4_cap_idx_t thread, l4_umword_t id)
{
  long err;

  *ref = ipc_cap_alloc();

  if (l4_is_invalid_cap(*ref))
    return -L4_ENOENT;

  /* Associate the object with an IPC gate. */

  err = l4_error(l4_factory_create_gate(l4re_env()->factory, *ref, thread, id));

  if (err)
    ipc_cap_free_um(*ref);

  return err;
}



/* Support for the convenience macro starting a server for a given object. */

long _ipc_server_loop_for(int expected_items, void *handler_obj,
                          ipc_server_handler_type handler, const char *name)
{
  ipc_server_config_type config;
  long err;

  if (name != NULL)
  {
    err = ipc_server_bind(name, (l4_umword_t) &config, &config.server);
    if (err)
      return err;
  }

  _ipc_server_init_for(&config, expected_items, handler_obj, handler);

  return ipc_server_start_config(&config);
}

/* Initialise a server for a given object. */

long _ipc_server_init_for(ipc_server_config_type *config, int expected_items,
                          void *handler_obj, ipc_server_handler_type handler)
{
  ipc_server_init_config(config);

  config->expected_items = expected_items;
  config->handler = handler;
  config->handler_obj = handler_obj;
}

/* Associate a new configuration with an existing server endpoint. */

long _ipc_server_add_config(ipc_server_config_type *config, int expected_items,
                            void *handler_obj, ipc_server_handler_type handler,
                            l4_cap_idx_t thread)
{
  _ipc_server_init_for(config, expected_items, handler_obj, handler);
  return ipc_server_start_config_thread(config, thread);
}

/* Complete initialisation of a server in the given thread. */

long ipc_server_start_config_thread(ipc_server_config_type *config,
                                    l4_cap_idx_t thread)
{
  config->thread = thread;
  config->config_thread = 1;

  return ipc_server_start_config(config);
}



/* A server main loop handling endpoint deletion for an IPC gate dedicated to
   a particular object within its own thread. */

long ipc_server_managed_loop(int expected_items, ipc_server_config_type *config)
{
  ipc_message_t msg;
  l4_umword_t label, irq_label;

  /* Permit other endpoints by dynamically interpreting the label. */

  ipc_server_config_type *config_from_label;

  /* Declare the extent to which capabilities are expected in messages. */

  long err = ipc_message_expect(&msg, expected_items);

  if (err)
    return err;

  /* Wait for an incoming message. */

  while (1)
  {
    ipc_message_wait(&msg, &label);

    /* Clear lower label bits. */

    label = label & ~3UL;

    /* Ignore erroneous messages. */

    if (l4_ipc_error(msg.tag, l4_utcb()))
      continue;

    /* Message involves the IPC gate itself. */

    irq_label = (l4_umword_t) config->irq;

    if (!config->notifications || (config->notifications && (label != irq_label)))
    {
      config_from_label = (ipc_server_config_type *) label;
      config_from_label->handler(&msg, config_from_label->handler_obj);
    }

    /* Message involves the IRQ or a termination condition occurred. */

    else if ((config->notifications && (label == irq_label)) || msg.terminating)
      break;
  }

  /* Free expected capabilities. */

  ipc_message_free(&msg);

  return L4_EOK;
}

/* A pthread-compatible mainloop initiation function. */

void *ipc_server_start_mainloop(void *data)
{
  ipc_server_config_type *config = (ipc_server_config_type *) data;

  long err = ipc_server_managed_loop(config->expected_items, config);

  if (config->finaliser != NULL)
    config->finaliser(config);

  return (void *) err;
}

/* Wait for an incoming message via an IPC gate dedicated to a particular
   object within its own thread. */

l4_msgtag_t ipc_server_wait(l4_umword_t id)
{
  l4_umword_t label;
  l4_msgtag_t tag;

  /* Ignore errors and messages not intended for the identified gate. */

  do tag = l4_ipc_wait(l4_utcb(), &label, L4_IPC_NEVER);
  while (l4_ipc_error(tag, l4_utcb()) || (label & ~3UL) != id);

  return tag;
}



/* Finalise a server configuration. */

void ipc_server_finalise_config(ipc_server_config_type *config)
{
  if (l4_is_valid_cap(config->server))
  {
    ipc_cap_free_um(config->server);
    config->server = L4_INVALID_CAP;
  }

  if (l4_is_valid_cap(config->irq))
  {
    ipc_cap_free_um(config->irq);
    config->irq = L4_INVALID_CAP;
  }
}

/* Discard any uninitiated thread. */

void ipc_server_discard_thread(ipc_server_config_type *config)
{
  if (config->config_thread && l4_is_valid_cap(config->thread))
  {
    ipc_cap_free_um(config->thread);
    config->thread = L4_INVALID_CAP;
  }
}

/* Initialise a server configuration. */

void ipc_server_init_config(ipc_server_config_type *config)
{
  /* No object references defined. */

  config->handler_obj = NULL;
  config->finaliser_obj = NULL;

  /* Expect no capability items by default. */

  config->expected_items = 0;

  /* No handler or finaliser defined. */

  config->handler = NULL;
  config->finaliser = NULL;

  /* No separate thread for the main loop. */

  config->config_thread = 0;

  /* Main thread by default with IPC gate to be allocated. */

  config->thread = l4re_env()->main_thread;
  config->server = L4_INVALID_CAP;

  /* No notifications and with IRQ to be potentially allocated. */

  config->notifications = 0;
  config->irq = L4_INVALID_CAP;
}

/* Initialise and start a server using the given configuration. */

long ipc_server_start_config(ipc_server_config_type *config)
{
  long err;

  /* Allocate a new IPC gate if one is not already defined. */

  if (l4_is_invalid_cap(config->server))
  {
    err = ipc_server_new_for_thread(&config->server, config, config->thread);

    if (err)
      return err;
  }

  /* Allocate an IRQ for notifications if requested. */

  if (config->notifications)
  {
    err = ipc_server_apply_irq_for_thread(config->server, &config->irq, config->thread);

    if (err)
      return err;

    /* Unmask the interrupt. */

    ipc_init_irq(config->irq);
  }

  /* With a separate thread, return the last status value. Otherwise, invoke the
     main loop. */

  if (config->config_thread)
    return L4_EOK;
  else
    return (long) ipc_server_start_mainloop(config);
}

/* vim: tabstop=2 expandtab shiftwidth=2
*/
